/* GLIB - Library of useful routines for C programming
 * Copyright (C) 1995-1997  Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * gthread.c: posix thread system implementation
 * Copyright 1998 Sebastian Wilhelmi; University of Karlsruhe
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * Modified by the GLib Team and others 1997-1999.  See the AUTHORS
 * file for a list of people on the GLib Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GLib at ftp://ftp.gtk.org/pub/gtk/.
 */

/* HISTORY
 *
 * 6th July 2000
 * almost the entire file rewritten following code from
 * Carlos hasan <chasan@acm.org>.  This file now uses native BeOS TLS
 * and benaphores.
 *
 * 3rd July 2000
 * First pass at getting the function implemented on BeOS.  This
 * isn't pretty but builds and sems to pass all the tests.  It should
 * be rewritten to use benaphores for greater speed and the TLS stuff
 * should be revisited.
 * David Reid <dreid@jetnet.co.uk>
 */

/* these are provided by BeOS, but don't normally get included.
 * undefine them here for ease of use.
 */
#undef MIN
#undef MAX
#include <kernel/OS.h>
#include <support/TLS.h>
#include <errno.h>
#include <stdlib.h>

#define BE_MAX_PRIVATE_KEYS 256

#define beos_print_error( name, num )                          \
  g_error( "file %s: line %d (%s): error %s during %s",         \
           __FILE__, __LINE__, G_GNUC_PRETTY_FUNCTION,          \
           g_strerror((num)), #name )


typedef struct _GMutex {
    sem_id		sem;
    vint32		count;
    thread_id	        owner;
} be_mutex_t;

#define G_MUTEX_SIZE (sizeof(struct _GMutex))

static GMutex *
g_mutex_new_beos_impl (void)
{
  GMutex *mutex = (GMutex *) g_new(be_mutex_t, 1);
  if (mutex != NULL) {
    mutex->sem = create_sem(0, "glib_mutex");
    mutex->count = 0;
    mutex->owner = 0;
    if (mutex->sem < B_OK) {
      g_free(mutex);
      mutex = NULL;
    }
  }

  return mutex;
}

static void
g_mutex_free_beos_impl (GMutex * mutex)
{
  if (mutex != NULL) {
    delete_sem(mutex->sem);
    g_free(mutex);
  }
}

/* NOTE: the functions g_mutex_lock and g_mutex_unlock may not use
   functions from gmem.c and gmessages.c; */

static void
g_mutex_lock_beos_impl (GMutex * mutex)
{
  thread_id caller = find_thread(NULL); /* so we only call it once */
  if (mutex != NULL)
    /* if we already own it why lock it again? pthreads says we should
     * return EDEADLK in this case, but here we just ignore it as
     * we don't return a value.
     */
    if (mutex->owner != caller) {
      if (atomic_add(&mutex->count, 1) != 0)
        acquire_sem(mutex->sem);
      mutex->owner = caller;
    }
}

static void
g_mutex_unlock_beos_impl(GMutex * mutex)
{
  if (mutex == NULL)
    return;
    /* if we don't own it we can't unlock it.  pthreads says we should
     * return EPERM in this case but as we don't return a vlue just
     * ignore it.
     */
  if (mutex->owner == find_thread(NULL)) {
    mutex->owner = 0;
    if (atomic_add(&mutex->count, -1) != 1)
      release_sem_etc(mutex->sem, 1, B_DO_NOT_RESCHEDULE);
  }
}

/* Programming with POSIX Threads has the following definition of
 * pthread_mutex_trylock()
 *
 * Lock a mutex.  If the mutex is currently locked, returns immeadiately
 * with EBUSY. Otherwise, calling thread becomes owner until it unlocks.
 *
 * Try to implememt that logic here.
 */
static gboolean
g_mutex_trylock_beos_impl (GMutex * mutex)
{
  if (mutex == NULL)
    return FALSE;

  /* there's no return value from g_mutex_lock_beos_impl so
   * we need to do our checking here!
   */

  /* Are we the owner of the mutex already?  If we do then
   * pthreads says we should return EDEADLK but as we only return
   * TRUE/FALSE what should we return?
   * The posix implementation would appear to return true so we'll
   * do the same.  This might not be right!
   */
  if (mutex->owner == find_thread(NULL))
    return TRUE;

  /* if the count != 0 then we've already got at least one lock
   * on the mutex so we shoudl return EBUSY...or in this case FALSE
   */
  if (mutex->count != 0)
    return FALSE;

  /*
   * potential deadlock here! 
   *
   * Carlos explains - if while the thread calling the trylock is here,
   * another thread grabs the lock we've got a lock and have returned
   * in error.  
   */
  g_mutex_lock_beos_impl(mutex);

  return TRUE;
}

typedef struct _GCond {
	sem_id		lock;
	sem_id		wait;
	sem_id		signal;
    vint32		waiters;
    vint32		ns;
} be_cond_t;

static GCond *
g_cond_new_beos_impl (void)
{
  GCond *cond = g_new(be_cond_t ,1);

  if (cond != NULL) {
  	cond->lock = create_sem(1, "glib_cond_lock");
  	cond->wait = create_sem(0, "glib_cond_wait");
  	cond->signal = create_sem(0, "glib_cond_signal");
  	if (cond->lock < B_OK || cond->wait < B_OK || cond->signal < B_OK) {
  	  delete_sem(cond->lock);
  	  delete_sem(cond->wait);
  	  delete_sem(cond->signal);
  	  g_free(cond);
  	  cond = NULL;
  	}
  }
  return cond;
}

static void
g_cond_free_beos_impl(GCond *cond)
{
  if (cond != NULL) {
    delete_sem (cond->lock);
    delete_sem (cond->wait);
    delete_sem (cond->signal);
    g_free (cond);
  }
}

static void
g_cond_wait_beos_impl (GCond *cond, GMutex *mutex)
{
  if (cond != NULL && mutex !=NULL) {   
    if (acquire_sem(cond->lock) == B_INTERRUPTED) return;
    cond->waiters++;
    release_sem(cond->lock);
    
    g_mutex_unlock_beos_impl (mutex);
    
    acquire_sem(cond->wait);

    while (acquire_sem(cond->lock) == B_INTERRUPTED){}
    if (cond->ns > 0){
      release_sem_etc(cond->signal, 1, B_DO_NOT_RESCHEDULE);
      cond->ns--;
    }        
    cond->waiters--;
    release_sem_etc(cond->lock, 1, B_DO_NOT_RESCHEDULE);

    g_mutex_lock_beos_impl (mutex);
  }
}

static gboolean
g_cond_timed_wait_beos_impl( GCond *cond, GMutex *mutex, GTimeVal *abs_time)
{
  status_t status = B_ERROR;

  if (cond != NULL && mutex != NULL && abs_time != NULL) {
    if (acquire_sem(cond->lock) == B_INTERRUPTED) return FALSE;
    cond->waiters++;
    release_sem(cond->lock);

    g_mutex_unlock_beos_impl (mutex);

    status = acquire_sem_etc(cond->wait, 1, B_ABSOLUTE_TIMEOUT,
        abs_time->tv_usec + 1000000LL * abs_time->tv_sec);

    while (acquire_sem(cond->signal) == B_INTERRUPTED){}
    if (cond->ns > 0){
      release_sem_etc(cond->signal, 1, B_DO_NOT_RESCHEDULE);
      cond->ns--;
    }        
    cond->waiters--;
    release_sem_etc(cond->signal, 1, B_DO_NOT_RESCHEDULE);

    g_mutex_lock_beos_impl (mutex);
  }
  return (status == B_OK);
}

static void
g_cond_signal_beos_impl (GCond *cond)
{
  if (cond == NULL)
    return;

  if (acquire_sem(cond->lock) == B_INTERRUPTED) return;
  if (cond->waiters > cond->ns) {
    cond->ns++;
    release_sem(cond->wait);
    release_sem(cond->lock);
    while (acquire_sem(cond->signal) == B_INTERRUPTED){}
   } else {
     release_sem(cond->lock);
   }
}

static void
g_cond_broadcast_beos_impl (GCond *cond)
{
  int32 count;

  if (cond ==NULL)
    return;

  if (acquire_sem(cond->lock) == B_INTERRUPTED) return;
  if (cond->waiters > cond->ns) {
    count = cond->waiters - cond->ns;
    cond->ns = cond->waiters;
    release_sem_etc(cond->wait, count, B_DO_NOT_RESCHEDULE);
    release_sem(cond->lock);
    while (acquire_sem_etc(cond->signal, count, 0 ,0) == B_INTERRUPTED){}
  } else {
    release_sem(cond->lock);
  }
}



typedef struct _GPrivate {
  int32 tls;
  GDestroyNotify destructor;
} be_private_t;

static vint32 be_private_array_index = 0;
static vint32 be_private_array_len = 0;
static be_private_t be_private_array[BE_MAX_PRIVATE_KEYS];

static GPrivate *
g_private_new_beos_impl (GDestroyNotify destructor)
{
  be_private_t *private_key;
  int32 index;

  if ((index = atomic_add(&be_private_array_index, 1)) >= BE_MAX_PRIVATE_KEYS) {
  	atomic_add(&be_private_array_index, -1);
  	beos_print_error("g_private_new_beos_impl", B_NO_MEMORY);
  }

  private_key = &be_private_array[index];
  private_key->tls = tls_allocate();
  private_key->destructor = destructor;
  tls_set(private_key->tls, 0);

  while (index != be_private_array_len)
    snooze(0);
  atomic_add(&be_private_array_len, 1);

  return private_key;
}

static gpointer
g_private_get_beos_impl (GPrivate * private_key)
{
  if (private_key != NULL)
    return (gpointer) tls_get(private_key->tls);
  return NULL;
}

static void
g_private_set_beos_impl (GPrivate * private_key, gpointer value)
{
  if (private_key != NULL)
    tls_set(private_key->tls, value);
}



typedef struct {
  GThreadFunc thread_func;
  gpointer arg;
} g_thread_thunk_data;

static void
g_thread_exit_thunk_func (void *data)
{
  g_thread_thunk_data *thunk_data = (g_thread_thunk_data *) data;
  gpointer self = g_thread_self();
  int32 i;

  if (thunk_data != NULL) {
    for (i = 0; i < be_private_array_len; i++) {
      be_private_t *priv = &be_private_array[i];
      if (priv->destructor != NULL && g_private_get_beos_impl(priv) != NULL)
        priv->destructor(self);
    }
    g_free(thunk_data);
  }
}

static int32
g_thread_init_thunk_func (void *data)
{
  g_thread_thunk_data *thunk_data = (g_thread_thunk_data *) data;

  if (thunk_data != NULL) {
    on_exit_thread(g_thread_exit_thunk_func, thunk_data);
    if (thunk_data->thread_func != NULL)
      thunk_data->thread_func(thunk_data->arg);
  }
  return 0;
}

static void
g_thread_create_beos_impl (GThreadFunc thread_func,
			    gpointer arg,
			    gulong stack_size,
			    gboolean joinable,
			    gboolean bound,
			    GThreadPriority priority,
			    gpointer thread)
{
  thread_id tid;
  int32 thread_priority;
  g_thread_thunk_data *thunk_data;

  g_return_if_fail (thread_func);

  switch (priority) {
    case G_THREAD_PRIORITY_LOW:
      thread_priority = B_LOW_PRIORITY;
      break;
    case G_THREAD_PRIORITY_NORMAL:
      thread_priority = B_NORMAL_PRIORITY;
      break;
    case G_THREAD_PRIORITY_HIGH:
      thread_priority = B_DISPLAY_PRIORITY;
      break;
    case G_THREAD_PRIORITY_URGENT:
      thread_priority = B_URGENT_DISPLAY_PRIORITY;
      break;
    default:
      thread_priority = B_NORMAL_PRIORITY;
      break;
  }

  *((thread_id *) thread) = 0;

  thunk_data = g_new(g_thread_thunk_data, 1);
  if (thunk_data == NULL)
    beos_print_error("g_thread_create_beos_impl", B_NO_MEMORY);

  thunk_data->thread_func = thread_func;
  thunk_data->arg = arg;

  tid = spawn_thread(g_thread_init_thunk_func, "glib_thread",
    		thread_priority, thunk_data);
  if (tid < B_OK)
    beos_print_error("spawn_thread", tid);
  resume_thread(tid);

  *((thread_id *) thread) = tid;
}

static void
g_thread_yield_beos_impl (void)
{
  snooze(0);
}

static void
g_thread_join_beos_impl (gpointer thread)
{
  status_t exitval;
  /* This I don't like but if we don't ignore the errors then
   * this will virtually always fail the test programs due to the
   * short lifespan of the test threads.  If someone has a better
   * way of doing this please tell me!
   */
  wait_for_thread(*(thread_id*)thread, &exitval);
}

static void
g_thread_exit_beos_impl (void)
{
  exit_thread(0);
}

static void
g_thread_set_priority_beos_impl (gpointer thread, GThreadPriority priority)
{
  int32 new_thread_priority;
  status_t rv; /* we need this to get the correct error */

  switch(priority){
    case G_THREAD_PRIORITY_LOW:
    	new_thread_priority = B_LOW_PRIORITY; break;
    default: case G_THREAD_PRIORITY_NORMAL:
    	new_thread_priority = B_NORMAL_PRIORITY; break;
    case G_THREAD_PRIORITY_HIGH:
    	new_thread_priority = B_DISPLAY_PRIORITY; break;
    case G_THREAD_PRIORITY_URGENT:
    	new_thread_priority = B_URGENT_DISPLAY_PRIORITY; break;
    }

	if ((rv = set_thread_priority(*(thread_id*)thread, new_thread_priority)) < 0)
	    beos_print_error("set_thread_priority", rv);
}


static void
g_thread_self_beos_impl (gpointer thread)
{
  *(thread_id*)thread = find_thread(NULL);
}

static GThreadFunctions g_thread_functions_for_glib_use_default =
{
  g_mutex_new_beos_impl,
  g_mutex_lock_beos_impl,
  g_mutex_trylock_beos_impl,
  g_mutex_unlock_beos_impl,
  g_mutex_free_beos_impl,

  g_cond_new_beos_impl,
  g_cond_signal_beos_impl,
  g_cond_broadcast_beos_impl,
  g_cond_wait_beos_impl,
  g_cond_timed_wait_beos_impl,
  g_cond_free_beos_impl,

  g_private_new_beos_impl,
  g_private_get_beos_impl,
  g_private_set_beos_impl,

  g_thread_create_beos_impl,
  g_thread_yield_beos_impl,
  g_thread_join_beos_impl,
  g_thread_exit_beos_impl,
  g_thread_set_priority_beos_impl,
  g_thread_self_beos_impl

};
